#include <fstream>
#include <future>
#include <iostream>
#include <string>
#include <vector>
#include <array>
#include <chrono>
#include <math.h>
#include <random>
#include <cstring>
#include <stdexcept>

/**
 * \brief Return index for discretization of a continuous-time vector
 *
 * \param time_vector Vector to discretize along
 * \param t_step Time step for discretization
 * \param N Number of time steps
 *
 * \return Discretized vector
 */
std::vector<int> discretization_idx(const std::vector<double>& time_vector, const double t_step, const int N) {

    // Make sure that for each discretized time step i*t_step we can find an index idx
    // such that time[idx] <= i*t_step and time[idx+1] > i*t_step.
    assert(time_vector[0] <= 0 && time_vector.back() > (N - 1) * t_step);

    std::vector<int> arr;

    int idx = 0;
    for (int i = 0; i < N; ++i)
    {
        double t_val = i * t_step;

        // Find the last reaction time not later that t_val. This is at idx.
        while (idx + 1 < time_vector.size() && time_vector[idx + 1] <= t_val) {
            ++idx;
        }
        arr.push_back(idx);
    }
    return arr;
}

/**
 * \brief Binding-unbinding process implemented using Gillpie's stochastic simulation algorithm
 *
 * \param init_state Initial state of the binding-unbinding process
 * \param p_bind Binding rate
 * \param p_unbind Unbinding rate
 * \param volume Scaling factor
 * \param T_max Time horizon
 * \param t_step Time step
 * \param seed Seed
 *
 * \return Array of the form (X, t, X_t), where X is the discretized state trajectory, t the time vector (discretized) and X_t the time-derivative of X
 */
std::tuple<std::vector<std::array<double, 3>>, std::vector<double>, std::vector<std::array<double, 3>>> bindUnbind(const std::array<int, 3>& init_state, double p_bind, double p_unbind, const double volume, const double T_max, const double t_step, int seed) {

    // Initialize state
    std::array<double, 3> state;

    // Perform the type conversion element by element
    for (size_t i = 0; i < init_state.size(); i++)
    {
        state[i] = static_cast<double>(init_state[i]);
    }

    // Random number generator
    std::mt19937 mt;
    if (seed == -1)
    {
        std::random_device rd;
        mt = std::mt19937(rd());  // Initialize mt using random_device
    }
    else
    {
        mt = std::mt19937(seed); // Initialize mt using the specified seed for correlated sampling
    }

    std::uniform_real_distribution<double> uni_distr(0, 1);

    // Continuous-time trajectory
    std::vector<std::array<double, 3>> trajectory;
    trajectory.push_back(state);

    // Continuous-time trajectory_dt (time-derivative)
    std::vector<std::array<double, 3>> trajectory_dt;
    double S0_dt = p_unbind * state[2] - p_bind / volume * state[0] * state[1];
    double S1_dt = p_unbind * state[2] - p_bind / volume * state[0] * state[1];
    double S2_dt = p_bind / volume * state[0] * state[1] - p_unbind * state[2];
    std::array<double, 3> state_dt = {S0_dt, S1_dt, S2_dt};
    trajectory_dt.push_back(state_dt);

    // Continuous-time time vector
    double time = 0.0;
    std::vector<double> time_vector;
    time_vector.push_back(time);

    // Initialize alpha
    std::array<double, 2> alpha;

    while (time < T_max) {
        // Binding S0 + S1 -> S2
        alpha[0] = p_bind / volume * state[0] * state[1];
        // Unbinding S2 - > S0 + S1
        alpha[1] = p_unbind * state[2];

        double lambda = alpha[0] + alpha[1];

        if (lambda == 0)
        {
            break;
        }

        // Determine reaction time
        double p = uni_distr(mt);
        double tau = 1 / lambda * log(1 / p);

        // Increment of time
        time = time + tau;
        time_vector.push_back(time);

        // Probability of binding is given by alpha[0] / lambda, then output is true (= 1)
        std::bernoulli_distribution ber_dist(alpha[0] / lambda);
        bool reaction = ber_dist(mt);

        // Update state
        state[0] = state[0] - 2.0 * reaction + 1.0;
        state[1] = state[1] - 2.0 * reaction + 1.0;
        state[2] = state[2] + 2.0 * reaction - 1.0;

        // Save result
        trajectory.push_back(state);

        // Compute time derivative
        S0_dt = p_unbind * state[2] - p_bind / volume * state[0] * state[1];
        S1_dt = p_unbind * state[2] - p_bind / volume * state[0] * state[1];
        S2_dt = p_bind / volume * state[0] * state[1] - p_unbind * state[2];
        state_dt = {S0_dt, S1_dt, S2_dt};
        trajectory_dt.push_back(state_dt);
    }

    // Initialize discretized arrays
    const int num_steps = T_max / t_step;
    std::vector<std::array<double, 3>> trajectory_discretized;
    std::vector<std::array<double, 3>> trajectory_dt_discretized;
    std::vector<double> time_discretized;

    // Get time discretization index
    auto discrete_idx = discretization_idx(time_vector, t_step, num_steps);

    // Discretize continous-time trajectories given the time discretization index
    for (auto const& idx : discrete_idx) {
        trajectory_discretized.push_back(trajectory[idx]);
        trajectory_dt_discretized.push_back(trajectory_dt[idx]);
        time_discretized.push_back(time_vector[idx]);
    }

    return {trajectory_discretized, time_discretized, trajectory_dt_discretized};
}

/**
 * \brief Computes a binding-unbinding process
 *
 * \param MC Number of samples for the estimates of mean and variance
 * \param init_state Initial state of the binding-unbinding process
 * \param gamma Transition rate constants for birth and death
 * \param volume Scaling factor
 * \param T_max Time horizon
 * \param t_step Time step
 * \param h Finite differences step size
 * \param correlated_sampling If true, then correlated_sampling is active
 *
 * \return A pair of arrays (mean, var) of sample mean and sample variance.
 *  The arrays are of the form[X, dX/d_bind, dX/d_unbind, dt/d_bind,
 *  dt/d_unbind, dX/dt] where X denotes the trajectory of the state, and the
 *  rest are partial derivatives wrt the rates and time.
 */
std::array<std::vector<std::array<double, 14>>, 2> runbindUnbind(const int MC, const std::array<int, 3>& init_state, const std::array<double, 2>& gamma, const double volume, const double T_max, const double h, const double t_step, const bool correlated_sampling) {

    const int m = 8;  // Number of parallel tasks
    const int num_cols = 14;  // Number of columns in output
    const int num_steps = T_max / t_step;  // Number of time steps

    // Initialize matrix
    using ResultMatrix = std::vector<std::array<double, num_cols>>;
    ResultMatrix mean(num_steps, std::array<double, num_cols> {});
    ResultMatrix variance = mean;

    // Initialize matrix
    std::vector<ResultMatrix> results(MC, mean);

    // Lambda function for parallel Monte Carlo simulation
    auto evaluate = [&](int first, int last) {
        for (int i = first; i < last; ++i)
        {
            int seed = correlated_sampling == true ? i : -1;
            for (int d = 0; d < 2; ++d)  // Two directional derivatives
            {
                double p_bind = gamma[0];
                double p_unbind = gamma[1];

                auto [result, time_result, result_dt] = bindUnbind(init_state, p_bind, p_unbind, volume, T_max, t_step, -1);
                assert(result.size() == num_steps);
                assert(result_dt.size() == num_steps);
                assert(time_result.size() == num_steps);

                auto [result_forward, time_result_forward, result_dt_forward] = bindUnbind(init_state, d == 0 ? p_bind + h / 2.0 : p_bind, d == 1 ? p_unbind + h / 2.0 : p_unbind, volume, T_max, t_step, seed);
                assert(result_forward.size() == num_steps);
                assert(result_dt_forward.size() == num_steps);
                assert(time_result_forward.size() == num_steps);

                auto [result_backward, time_result_backward, result_dt_backward] = bindUnbind(init_state, d == 0 ? p_bind - h / 2.0 : p_bind, d == 1 ? p_unbind - h / 2.0 : p_unbind, volume, T_max, t_step, seed);
                assert(result_backward.size() == num_steps);
                assert(result_dt_backward.size() == num_steps);
                assert(time_result_backward.size() == num_steps);

                for (int j = 0; j < num_steps; ++j)
                {
                    results[i][j][0] += result[j][0] / 2;  // Compensate for adding up for two directions
                    results[i][j][1] += result[j][1] / 2;  // Compensate for adding up for two directions
                    results[i][j][2] += result[j][2] / 2;  // Compensate for adding up for two directions
                    results[i][j][3 + 3 * d] += (result_forward[j][0] - result_backward[j][0]) / h;
                    results[i][j][4 + 3 * d] += (result_forward[j][1] - result_backward[j][1]) / h;
                    results[i][j][5 + 3 * d] += (result_forward[j][2] - result_backward[j][2]) / h;
                    results[i][j][9 + d] += (time_result_forward[j] - time_result_backward[j]) / h;
                    results[i][j][11] += result_dt[j][0] / 2;  // Compensate for adding up for two directions
                    results[i][j][12] += result_dt[j][1] / 2;  // Compensate for adding up for two directions
                    results[i][j][13] += result_dt[j][2] / 2;  // Compensate for adding up for two directions
                }
            }
        }
    };

    // Parallel execution
    std::vector<std::future<void>> handles;
    for (int i = 0; i < m; ++i) {
        handles.push_back(std::async(std::launch::async, evaluate, i * MC / m, (i + 1) * MC / m));
    }
    for (auto& f : handles) {
        f.wait();
    }

    // Calculate row mean of results (each column contains a sample)
    for (auto const& result : results)
    {
        // Iterate through time
        for (int i = 0; i < result.size(); ++i)
        {
            // Iterate through values
            for (int j = 0; j < result[j].size(); ++j) {
                mean[i][j] += result[i][j];
            }
        }
    }
    for (auto& row : mean)
    {
        for (auto& val : row) {
            val /= MC;
        }
    }

    // Calculate row variance of results (each column contains a sample)
    for (auto const& result : results)
    {
        // Iterate through rows -> time
        for (int i = 0; i < result.size(); ++i)
        {
            // Iterate through columns -> values
            for (int j = 0; j < result[j].size(); ++j) {
                variance[i][j] += (result[i][j] - mean[i][j]) * (result[i][j] - mean[i][j]);
            }
        }
    }
    for (auto& row : variance)
    {
        for (auto& val : row) {
            val /= MC;
        }
    }

    return {mean, variance};
}

int main(int argc, char* argv[]) {
    const std::array<int, 3> init_state = {100, 100, 0};  // Initial state
    const double volume = 100;  // Scaling factor
    const double T_max = 25;  // Time horizon for simulation
    const int num_steps = 500;  // Number of time steps
    const int MC = 1e5;  // Number of Monte Carlo samples
    const double h = 0.01;  // Finite difference step size
    bool correlated_sampling = false;  // Correlated sampling
    const std::array<double, 2> gamma = {0.1, 0.2};  // Binding and unbinding rates

    try {
        if (argc == 1)
        {
            std::cout << "Run program without correlated sampling" << std::endl;
        }
        else if (argc == 2)
        {
            if (std::strcmp(argv[1], "--correlated-sampling") == 0)
            {
                std::cout << "Run program with correlated sampling for variance reduction" << std::endl;
                correlated_sampling = true;
            }
            else if (std::strcmp(argv[1], "-h") == 0 || std::strcmp(argv[1], "--h") == 0)
            {
                std::cout << "Available options:\n--correlated-sampling: correlated sampling for variance reduction" << std::endl;
                return 0;
            }
            else
            {
                throw std::invalid_argument("Invalid argument: " + std::string(argv[1]) + ". Did you mean \"--correlated-sampling\"?");
            }
        }
        else
        {
            throw std::invalid_argument("Invalid number of arguments");
        }
    } catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        return 1; // Return an error code
    }

    // Run simulation
    const double t_step = T_max / num_steps;
    auto tic = std::chrono::high_resolution_clock::now();
    auto [mean, variance] = runbindUnbind(MC, init_state, gamma, volume, T_max, h, t_step, correlated_sampling);
    auto toc = std::chrono::high_resolution_clock::now();

    std::ofstream outfile_1("out_bindUnbind_mean_FD.csv");
    for (const auto& row : mean) {
        for (const auto& val : row) {
            outfile_1 << val << " ";
        }
        outfile_1 << std::endl;
    }
    outfile_1.close();

    std::ofstream outfile_2("out_bindUnbind_var_FD.csv");
    for (const auto& row : variance) {
        for (const auto& val : row) {
            outfile_2 << val << " ";
        }
        outfile_2 << std::endl;
    }
    outfile_2.close();

    auto duration = std::chrono::duration_cast<std::chrono::seconds>(toc - tic);
    std::cout << "\nElapsed time: " << duration.count() << " seconds" << std::endl;

    return 0;
}