#include "sad.hh"

#include <fstream>
#include <iostream>
#include <string>
#include <vector>

using namespace StochasticDifferentiation;
using Real = SAD<double>;

/**
 * \brief Compute the state of a binding-unbinding process for infinitly many molecules
 *
 * \param init_state Initial state of the binding-unbinding process
 * \param p_bind Binding rate
 * \param p_unbind Unbinding rate
 *
 * \return New state of the binding-unbinding process
 */
std::array<Real, 3> bindUnbind(const std::array<Real, 3>& init_state, Real& p_bind, Real& p_unbind) {

	// Initialize state
	std::array<Real, 3> state = init_state;

	state[0] = -1.0 * p_bind * init_state[0] * init_state[1] + p_unbind * init_state[2];
	state[1] = -1.0 * p_bind * init_state[0] * init_state[1] + p_unbind * init_state[2];
	state[2] = p_bind * init_state[0] * init_state[1] - 1.0 * p_unbind * init_state[2];

	return state;
}

/** Compute a (discrete) trajectory of the binding-unbinding process for infinitly many molecules using an Euler scheme
 * \param init_state Initial state of the binding-unbinding process
 * \param p_bind Binding rate
 * \param p_unbind Unbinding rate
 * \param T_max Time horizon
 * \param t_step Time step
 * 
 * \return Discrete state trajectory of binding-unbinding process
 */
std::vector<std::array<Real, 3>> Euler(const std::array<double, 3> init_state, Real p_bind, Real p_unbind, const double T_max, const double t_step) {

	// Initialize state
	Real S0(p_unbind.sp, init_state[0], 0);
	Real S1(p_unbind.sp, init_state[1], 0);
	Real S2(p_unbind.sp, init_state[2], 0);
	std::array<Real, 3> state = {S0, S1, S2};

	std::vector<std::array<Real, 3>> trajectory;
	trajectory.push_back(state);

	double time = 0;

	while (time <= T_max) {
		auto dt = bindUnbind(state, p_bind, p_unbind);

		for (int i = 0; i < 3; ++i)
		{
			state[i] = state[i] + t_step * dt[i];
		}
		
		trajectory.push_back(state);
		time += t_step;
	}

	return trajectory;
}

int main() {
	const std::array<double, 3> init_state = {1.0, 1.0, 0};  // Initial state
    const double volume = 100;  // Scaling factor
    const double T_max = 25;  // Time horizon for simulation
    const int num_steps = 500;  // Number of time steps
    
    // Necessary to use the stochastic automatic differentiation toolbox
    const double w = 0.5;
    StochasticProgram<double> sp;
    sp.jumpWeight = w;

    const double t_step = T_max / num_steps;

    // Initialize matrix
    using ResultMatrix = std::vector<std::array<double, 9>>;
	ResultMatrix mean(num_steps, std::array<double, 9>{});

    for (int d = 0; d < 2; ++d)
    {
	    Real p_bind(&sp, 0.1, d==0 ? 1.0 : 0.0);
	    Real p_unbind(&sp, 0.2, d==1 ? 1.0 : 0.0);
    	auto trajectory = Euler(init_state, p_bind, p_unbind, T_max, t_step);

    	for (int j = 0; j < num_steps; ++j)
        {
        	mean[j][0] += trajectory[j][0].value / 2;
	        mean[j][1] += trajectory[j][1].value / 2;
	        mean[j][2] += trajectory[j][2].value / 2;
	        mean[j][3 + 3 * d] += trajectory[j][0].derivative;
	        mean[j][4 + 3 * d] += trajectory[j][1].derivative;
	        mean[j][5 + 3 * d] += trajectory[j][2].derivative;
        }
    }

    std::ofstream outfile("out_bindUnbind_ODE.csv");
    for (const auto& row : mean) {
        for (const auto& val: row) {
            outfile << volume * val << " ";
        }
        outfile << std::endl;
    }
    outfile.close();

	return 0;
}