#include <fstream>
#include <future>
#include <iostream>
#include <string>
#include <vector>
#include <array>
#include <chrono>
#include <math.h>
#include <random>
#include <cstring>
#include <stdexcept>

/**
 * \brief Discretizes a continuous-time vector
 *
 * \param trajectory Vector to discretize
 * \param time_vector Vector to discretize along
 * \param t_step Time step for discretization
 * \param N Number of time steps
 *
 * \return Discretized vector
 */
template <typename T>
std::vector<T> discretize(const std::vector<T>& trajectory, const std::vector<double>& time_vector,
                          const double t_step, const int N) {

    // Make sure that for each discretized time step i*t_step we can find an index idx
    // such that time[idx] <= i*t_step and time[idx+1] > i*t_step.
    assert(time_vector[0] <= 0 && time_vector.back() > (N - 1) * t_step);

    std::vector<T> arr;

    int idx = 0;
    for (int i = 0; i < N; ++i)
    {
        double t_val = i * t_step;

        // Find the last reaction time not later that t_val. This is at idx.
        while (idx + 1 < time_vector.size() && time_vector[idx + 1] <= t_val) {
            ++idx;
        }
        arr.push_back(trajectory[idx]);
    }
    return arr;
}

/**
 * \brief Birth-death process implemented using Gillpie's algorithm
 *
 * \param init_state Initial state of the birth-death process
 * \param p_birth Birth rate
 * \param p_death Death rate
 * \param volume Scaling factor
 * \param T_max Time horizon
 * \param t_step Time step
 * \param seed Seed
 *
 * \return Array of the form (X, t, X_t), where X is the discretized state trajectory, t the time vector (discretized) and X_t the time-derivative of X
 */
std::array<std::vector<double>, 3> birthDeath(const int init_state, double p_birth, double p_death, const double volume, const double T_max, const double t_step, int seed) {

    // Initialize state
    int state = init_state;

    // Random number generator
    std::mt19937 mt;
    if (seed == -1)
    {
        std::random_device rd;
        mt = std::mt19937(rd());  // Initialize mt using random_device
    }
    else
    {
        mt = std::mt19937(seed); // Initialize mt using the specified seed for correlated sampling
    }

    std::uniform_real_distribution<double> uni_distr(0, 1);

    // Continuous-time trajectory
    std::vector<double> trajectory;
    trajectory.push_back(state);

    // Continuous-time trajectory_dt (time-derivative)
    std::vector<double> trajectory_dt;
    trajectory_dt.push_back(p_birth * volume - p_death * state);  // = Expected jump / expected jump time

    double time = 0.0;
    std::vector<double> time_vector;
    time_vector.push_back(time);

    // Initialize alpha
    std::array<double, 2> alpha;

    while (time < T_max) {
        // Degradation S1 -> {}
        alpha[0] = p_death * state;
        // Production {} -> S1
        alpha[1] = p_birth * volume;

        double lambda = alpha[0] + alpha[1];

        if (lambda == 0)
        {
            break;
        }

        // Determine reaction time
        double p = uni_distr(mt);
        double tau = 1 / lambda * log(1 / p);

        // Increment of time
        time = time + tau;
        time_vector.push_back(time);

        // Probability of degradation is given by alpha[0] / lambda, then output is true (= 1)
        std::bernoulli_distribution ber_dist(alpha[0] / lambda);
        bool reaction = ber_dist(mt);

        // Update state
        state = state - 2.0 * reaction + 1.0;

        trajectory.push_back(state);
        trajectory_dt.push_back(alpha[1] - alpha[0]);

    }

    // Discrete-time
    const int num_steps = T_max / t_step;
    auto trajectory_discretized = discretize(trajectory, time_vector, t_step, num_steps);
    auto trajectory_dt_discretized = discretize(trajectory_dt, time_vector, t_step, num_steps);
    auto time_discretized = discretize(time_vector, time_vector, t_step, num_steps);

    return {trajectory_discretized, time_discretized, trajectory_dt_discretized};
}

/**
 * \brief Computes a birth-death process
 *
 * \param MC Number of samples for the estimates of mean and variance
 * \param init_state Initial state of the birth-death process
 * \param gamma Transition rate constants for birth and death
 * \param volume Scaling factor
 * \param T_max Time horizon
 * \param w Probability of taking the jump branch in discontinuous distributions
 * \param t_step Time step
 * \param h Finite differences step size
 * \param correlated_sampling If true, then correlated_sampling is active
 *
 * \return A pair of arrays (mean, var) of sample mean and sample variance.
 *  The arrays are of the form[X, dX/d_birth, dX/d_death, dt/d_birth,
 *  dt/d_death, dX/dt] where X denotes the trajectory of the state, and the
 *  rest are partial derivatives wrt the rates and time.
 */
std::array<std::vector<std::array<double, 6>>, 2> runbirthDeath(const int MC, const int init_state, const std::array<double, 2>& gamma, const double volume, const double T_max, const double t_step, const double h, const bool correlated_sampling) {

    const int num_steps = T_max / t_step;  // Number of parallel tasks
    const int num_cols = 6;  // Number of columns in output
    const int m = 8;  // Number of parallel tasks

    using ResultMatrix = std::vector<std::array<double, num_cols>>;
    ResultMatrix mean(num_steps, std::array<double, num_cols> {});
    ResultMatrix variance = mean;

    std::vector<ResultMatrix> results(MC, mean);

    // Lambda function for parallel Monte Carlo simulation
    auto evaluate = [&](int first, int last) {
        for (int i = first; i < last; ++i)
        {
            int seed = correlated_sampling == true ? i : -1;
            for (int d = 0; d < 2; ++d)  // Two directional derivatives
            {
                auto [result, time_result, result_dt] = birthDeath(init_state, gamma[0], gamma[1], volume, T_max, t_step, -1);
                assert(result.size() == num_steps);
                assert(result_dt.size() == num_steps);
                assert(time_result.size() == num_steps);

                auto [result_forward, time_result_forward, result_dt_forward] = birthDeath(init_state, d == 0 ? gamma[0] + h / 2.0 : gamma[0], d == 1 ? gamma[1] + h / 2.0 : gamma[1], volume, T_max, t_step, seed);
                assert(result_forward.size() == num_steps);
                assert(result_dt_forward.size() == num_steps);
                assert(time_result_forward.size() == num_steps);

                auto [result_backward, time_result_backward, result_dt_backward] = birthDeath(init_state, d == 0 ? gamma[0] - h / 2.0 : gamma[0], d == 1 ? gamma[1] - h / 2.0 : gamma[1], volume, T_max, t_step, seed);
                assert(result_backward.size() == num_steps);
                assert(result_dt_backward.size() == num_steps);
                assert(time_result_backward.size() == num_steps);

                for (int j = 0; j < num_steps; ++j)
                {
                    results[i][j][0] += result[j] / 2;
                    results[i][j][1 + d] += (result_forward[j] - result_backward[j]) / h;  // Central finite differences
                    results[i][j][3 + d] += (time_result_forward[j] - time_result_backward[j]) / h;
                    results[i][j][5] += result_dt[j] / 2;  // Compensate for adding up for two directions
                }
            }
        }
    };

    // Parallel execution
    std::vector<std::future<void>> handles;
    for (int i = 0; i < m; ++i) {
        handles.push_back(std::async(std::launch::async, evaluate, i * MC / m, (i + 1) * MC / m));
    }
    for (auto& f : handles) {
        f.wait();
    }

    // Calculate row mean of results (each column contains a sample)
    for (auto const& each_matrix : results) {
        // Iterate through rows -> time
        for (int i = 0; i < each_matrix.size(); ++i)
        {
            // Iterate through columns -> values
            for (int j = 0; j < each_matrix[j].size(); ++j)
            {
                mean[i][j] += each_matrix[i][j];
            }
        }
    }
    for (auto& row : mean)
    {
        for (auto& val : row) {
            val /= MC;
        }
    }

    // Calculate row variance of results (each column contains a sample)
    for (auto const& each_matrix : results) {
        // Iterate through rows -> time
        for (int i = 0; i < each_matrix.size(); ++i)
        {
            // Iterate through columns -> values
            for (int j = 0; j < each_matrix[j].size(); ++j)
            {
                variance[i][j] += (each_matrix[i][j] - mean[i][j]) * (each_matrix[i][j] - mean[i][j]);
            }
        }
    }
    for (auto& row : variance)
    {
        for (auto& val : row) {
            val /= MC;
        }
    }

    return {mean, variance};
}

int main(int argc, char* argv[]) {

    const int init_state = 0;  // Initial state
    const double volume = 10;  // Scaling factor
    const double T_max = 100;  // Time horizon for simulation
    const int num_steps = 500;  // Number of time steps
    const double h = 0.01;  // Finite difference step size
    const int MC = 1e5;  // Number of Monte Carlo samples
    bool correlated_sampling = false;  // Correlated sampling
    const std::array<double, 2> gamma = {0.2, 0.1};  // Birth and death rates

    try {
        if (argc == 1)
        {
            std::cout << "Run program without correlated sampling" << std::endl;
        }
        else if (argc == 2)
        {
            if (std::strcmp(argv[1], "--correlated-sampling") == 0)
            {
                std::cout << "Run program with correlated sampling for variance reduction" << std::endl;
                correlated_sampling = true;
            }
            else if (std::strcmp(argv[1], "-h") == 0 || std::strcmp(argv[1], "--h") == 0)
            {
                std::cout << "Available options:\n--correlated-sampling: correlated sampling for variance reduction" << std::endl;
                return 0;
            }
            else
            {
                throw std::invalid_argument("Invalid argument: " + std::string(argv[1]) + ". Did you mean \"--correlated-sampling\"?");
            }
        }
        else
        {
            throw std::invalid_argument("Invalid number of arguments");
        }
    } catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        return 1; // Return an error code
    }

    const double t_step = T_max / num_steps;
    auto tic = std::chrono::high_resolution_clock::now();
    auto [mean, variance] = runbirthDeath(MC, init_state, gamma, volume, T_max, t_step, h, correlated_sampling);
    auto toc = std::chrono::high_resolution_clock::now();

    std::ofstream outfile_1("out_birthDeath_mean_FD.csv");
    for (const auto& row : mean) {
        for (const auto& val : row) {
            outfile_1 << val << " ";
        }
        outfile_1 << std::endl;
    }
    outfile_1.close();

    std::ofstream outfile_2("out_birthDeath_var_FD.csv");
    for (const auto& row : variance) {
        for (const auto& val : row) {
            outfile_2 << val << " ";
        }
        outfile_2 << std::endl;
    }
    outfile_2.close();

    auto duration = std::chrono::duration_cast<std::chrono::seconds>(toc - tic);
    std::cout << "\nElapsed time: " << duration.count() << " seconds" << std::endl;

    return 0;
}