#include "sad.hh"

#include <fstream>
#include <iostream>
#include <string>
#include <vector>

using namespace StochasticDifferentiation;
using Real = SAD<double>;

/**
 * \brief Compute the state trajectory of a birth-death process for infinitly many molecules using an Euler scheme
 *
 * \param init_state Initial state of the birth-death
 * \param p_birth Birth rate
 * \param p_death Death rate
 * \param T_max Time horizon
 * \param t_step Time step
 *
 * \return Discrete state trajectory of the birth-death process
 */
std::vector<Real> birthDeath(const int init_state, Real p_birth, Real p_death, const double T_max, const double t_step) {

	// Initialize state
	Real state(p_birth.sp, init_state, 0);

	std::vector<Real> trajectory;
	trajectory.push_back(state);

	double time = 0;

	while (time <= T_max) {
		state = state + t_step * (-1.0 * p_death * state + p_birth);
		trajectory.push_back(state);
		time += t_step;
	}

	return trajectory;
}

int main() {
	const int init_state = 0;  // Initial state
    const double volume = 10;  // Scaling factor
    const double T_max = 100;  // Time horizon for simulation
    const int num_steps = 500;  // Number of time steps
    
    // Necessary to use the stochastic automatic differentiation toolbox
    const double w = 0.01;
    StochasticProgram<double> sp;
    sp.jumpWeight = w;

	const double t_step = T_max / num_steps;

    using ResultMatrix = std::vector<std::array<double, 3>>;
	ResultMatrix mean(num_steps, std::array<double, 3>{});

    for (int d = 0; d < 2; ++d)
    {
	    Real p_birth(&sp, 0.2, d==0 ? 1.0 : 0.0);
	    Real p_death(&sp, 0.1, d==1 ? 1.0 : 0.0);
	    auto trajectory = birthDeath(init_state, p_birth, p_death, T_max, t_step);

    	for (int j = 0; j < num_steps; ++j)
        {
        	mean[j][0] += trajectory[j].value / 2;
	        mean[j][1 + d] += trajectory[j].derivative;
        }
    }

    std::ofstream outfile("out_birthDeath_ODE.csv");
    for (const auto& row : mean) {
        for (const auto& val: row) {
            outfile << volume * val << " ";
        }
        outfile << std::endl;
    }
    outfile.close();

	return 0;
}