#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug  3 10:36:24 2023

@author: Jan-Hendrik Niemann
"""


import numpy as np
from math import exp
import matplotlib.pyplot as plt


# Figure export
export_image = False

# %% Load data

data_mean = np.genfromtxt('out_birthDeath_mean.csv', delimiter=" ", dtype=float)
data_var = np.genfromtxt('out_birthDeath_var.csv', delimiter=" ", dtype=float)
data_std = np.sqrt(data_var)
data_mean_FD = np.genfromtxt('out_birthDeath_mean_FD.csv', delimiter=" ", dtype=float)
data_var_FD = np.genfromtxt('out_birthDeath_var_FD.csv', delimiter=" ", dtype=float)
data_std_FD = np.sqrt(data_var_FD)

# %% Simulation settings

T_max = 100
time = np.linspace(0, T_max, data_mean.shape[0])
p_death = 0.1
p_birth = 0.2
num_timesteps = data_mean.shape[0]
t_step = T_max / num_timesteps

# Acts as scaling factor. Scales each reactant equally
volume = 10


def BD_ODE(x_init, t_step, T_max, gamma):
    """
    Ordinary differential equation approximating the expectation for infinite
    volumes

    Parameters
    ----------
    x_init : ndarray
        Initial concentration.
    t_step : float
        Time step.
    T_max : int or float
        Time horizon.
    gamma : ndarray
        Matrix containing transition rate constants for degradation and
        production.

    Returns
    -------
    C : ndarray
        Trajectory of the system state for given time horizon.

    """

    num_timesteps = int(np.round(T_max / t_step))

    C = np.zeros(int(num_timesteps))
    C[0] = x_init

    for k in range(num_timesteps - 1):

        C[k + 1] = C[k] + t_step * (- gamma[0] * C[k] + gamma[1])

    return C


def func_death(t, p_birth, p_death, volume):
    p = p_birth / p_death
    return t * volume * p * exp(-p_death * t) - volume * p / p_death + volume * p / p_death * exp(-p_death * t)


def func_birth(t, p_death, volume):
    return volume / p_death * (1 - exp(- p_death * t))


y_death = np.zeros_like(time)
y_birth = np.zeros_like(time)
y_birth2 = np.zeros_like(time)
for i, t in enumerate(time):
    y_death[i] = func_death(t, p_birth, p_death, volume)
    y_birth[i] = func_birth(t, p_death, volume)

# %% Simulate and plot

# Transition rate constants death birth
gamma = np.array([p_death, p_birth], dtype=float)

# Initial state
C_init = 0.0

# Run simulation
C = BD_ODE(C_init, t_step, T_max, gamma)

fig = plt.figure()
plt.plot(np.linspace(0, T_max, num_timesteps), volume * C, label='Anal.', c='k')
plt.plot(time, data_mean[:, 0], label='SAD')
plt.fill_between(time, data_mean[:, 0] - data_std[:, 0], data_mean[:, 0] + data_std[:, 0], alpha=0.25)
plt.plot(time, data_mean_FD[:, 0], label='FD')
plt.fill_between(time, data_mean_FD[:, 0] - data_std_FD[:, 0], data_mean_FD[:, 0] + data_std_FD[:, 0], alpha=0.25)
plt.title('$E(t)$')
plt.xlabel('Time $t$')
plt.legend()

# Compute variance and standard deviation wrt death rate
var_1 = data_var[:, 1] + data_mean[:, 5]**2 * data_var[:, 3] + data_mean[:, 3]**2 * data_var[:, 5] + data_var[:, 5] * data_var[:, 3] - 2 * (data_mean[:, 6] - data_mean[:, 5] * data_mean[:, 3] * data_mean[:, 1])
std_1 = np.sqrt(var_1)

fig = plt.figure()
plt.plot(time, y_birth, label='Anal.', c='k')
plt.plot(time, data_mean[:, 1] - data_mean[:, 3] * data_mean[:, 5], label='SAD')
plt.fill_between(time, data_mean[:, 1] - data_mean[:, 3] * data_mean[:, 5] - std_1,
                 data_mean[:, 1] - data_mean[:, 3] * data_mean[:, 5] + std_1, alpha=0.25)
plt.plot(time, data_mean_FD[:, 1], label='FD')
plt.fill_between(time, data_mean_FD[:, 1] - data_std_FD[:, 1], data_mean_FD[:, 1] + data_std_FD[:, 1], alpha=0.25)
plt.title(r'$\frac{\partial}{\partial\gamma_2}E(t)$')
plt.legend()
plt.xlabel('Time $t$')

# Compute variance and standard deviation wrt birth rate
var_0 = data_var[:, 2] + data_mean[:, 5]**2 * data_var[:, 4] + data_mean[:, 4]**2 * data_var[:, 5] + data_var[:, 5] * data_var[:, 4] - 2 * (data_mean[:, 7] - data_mean[:, 5] * data_mean[:, 4] * data_mean[:, 2])
std_0 = np.sqrt(var_0)

fig = plt.figure()
plt.plot(time, y_death, label='Anal.', c='k')
plt.plot(time, data_mean[:, 2] - data_mean[:, 4] * data_mean[:, 5], label='SAD')
plt.fill_between(time, data_mean[:, 2] - data_mean[:, 4] * data_mean[:, 5] - std_0,
                 data_mean[:, 2] - data_mean[:, 4] * data_mean[:, 5] + std_0, alpha=0.25)
plt.plot(time, data_mean_FD[:, 2], label='FD')
plt.fill_between(time, data_mean_FD[:, 2] - data_std_FD[:, 2], data_mean_FD[:, 2] + data_std_FD[:, 2], alpha=0.25)
plt.title(r'$\frac{\partial}{\partial\gamma_1}E(t)$')
plt.xlabel('Time')
plt.legend()
plt.xlabel('Time $t$')

# %% Figure export

fig, axs = plt.subplots(2, sharex=True)

axs[0].plot(time, y_death, c='k', label='_nolegend_')
axs[0].plot(time, data_mean[:, 2] - data_mean[:, 4] * data_mean[:, 5], label='SAD')
axs[0].fill_between(time, data_mean[:, 2] - data_mean[:, 4] * data_mean[:, 5] - std_0, data_mean[:, 2] - data_mean[:, 4] * data_mean[:, 5] + std_0, alpha=0.25, label='_nolegend_')
axs[0].plot(time, data_mean_FD[:, 2], label='FD')
axs[0].fill_between(time, data_mean_FD[:, 2] - data_std_FD[:, 2], data_mean_FD[:, 2] + data_std_FD[:, 2], alpha=0.25, label='_nolegend_')
axs[0].set_title(r'$\frac{\partial}{\partial\gamma_1}E(p, t)$', pad=15)

axs[1].plot(time, y_birth, c='k', label='_nolegend_')
axs[1].plot(time, data_mean[:, 1] - data_mean[:, 3] * data_mean[:, 5], label='SAD')
axs[1].fill_between(time, data_mean[:, 1] - data_mean[:, 3] * data_mean[:, 5] - std_1, data_mean[:, 1] - data_mean[:, 3] * data_mean[:, 5] + std_1, alpha=0.25, label='_nolegend_')
axs[1].plot(time, data_mean_FD[:, 1], label='FD')
axs[1].fill_between(time, data_mean_FD[:, 1] - data_std_FD[:, 1], data_mean_FD[:, 1] + data_std_FD[:, 1], alpha=0.25, label='_nolegend_')
axs[1].set_title(r'$\frac{\partial}{\partial\gamma_2}E(p, t)$', pad=15)
plt.xlabel('Time $t$')
plt.figlegend(['SAD', 'FD'], ncols=2, loc=10, bbox_to_anchor=(0.535, 0.0))
fig.tight_layout()

if export_image:
    plt.savefig('BDP.png', bbox_inches='tight', transparent=True)
