#include "sad.hh"

#include <array>
#include <fstream>
#include <future>
#include <iostream>
#include <string>


using namespace StochasticDifferentiation;
using Real = SAD<double>;

std::array<Real,2> ipd(Real pa, Real qa, Real pb, Real qb, int n)
{
  // initialize payoff of both players to zero
  Real payoffA(pa.sp,0,0);
  Real payoffB(pa.sp,0,0);

  // initialize the players strategy as if the previous (non-existent)
  // round had been cooperation on both sides
  Real strategyA = pa;
  Real strategyB = pb;

  // play n rounds
  for (int i=0; i<n; ++i)
  {
    // the moves of the players (cooperation=1, betrayal=0) are just Bernoulli
    // realizations with success (cooperation) probability defined by the
    // respective player's strategy.
    Real moveA = bernoulli(strategyA);
    Real moveB = bernoulli(strategyB);

    // the payoff for each player depends on both player's moves. This follows the
    // classical prisoners' dilemma setup.
    payoffA = payoffA -2.0*moveA*moveB -10.0*moveA*(1.0-moveB) -5.0*(1.0-moveA)*(1.0-moveB);
    payoffB = payoffB -2.0*moveA*moveB -10.0*(1.0-moveA)*moveB -5.0*(1.0-moveA)*(1.0-moveB);

    // update the strategies based on the opponent's move: select cooperation probability
    // either p (opponent betrayed) or q (opponent collaborated).
    strategyA = pa*moveB + qa*(1.0-moveB);
    strategyB = pb*moveA + qb*(1.0-moveA);
  }

  return {payoffA,payoffB};
}

/**
 * \brief Computes the iterated prisoners' dilemma.
 *
 * \param p collaboration probability when opponent collaborated in the previous round
 * \param q collaboration probability when opponent betrayed in the previous round
 * \param n number of rounds to play
 * \param N number of samples to draw for each partial derivative (payoff values get four times the samples)
 * \param w probability of taking the jump branch in discontinuous distributions
 *
 * \return a pair of arrays (mean,var) of sample mean and sample variance. The arrays are of the form
 *         [a,b,da/dpa,db/dpa,da/dqa,db/dqa,da/dpb,db/dpb,da/dqb,db/dqb] where
 *         a denotes the payoff of player A, b the payoff of player B, and the rest are
 *         partial derivatives wrt the players' strategy parameters.
 */
// [a,b,da/dpa,db/dpa,da/dqa,db/dqa,da/dpb,db/dpb,da/dqb,db/dqb]
//  0 1   2      3      4      5      6      7      8      9
std::array<std::array<double,10>,2> runIpd(double p, double q, int n, int N, double w)
{
  using ResultArray = std::array<double,10>;
  ResultArray mean;
  std::fill(begin(mean),end(mean),0.0);
  ResultArray variance = mean;

  std::vector<ResultArray> results(N,mean);

  auto evaluate = [&](int first, int last)
  {
    for (int i=first; i<last; ++i)                     // drawing N samples,
    {
      for (int d=0; d<4; ++d)                   // compute directional derivatives in four different directions
      {
        StochasticProgram<double> sp;
        sp.jumpWeight = w;

        Real pa(&sp,p,d==0? 1.0: 0.0);          // directions are unit vectors in direction of strategy probabilities
        Real qa(&sp,q,d==1? 1.0: 0.0);
        Real pb(&sp,p,d==2? 1.0: 0.0);
        Real qb(&sp,q,d==3? 1.0: 0.0);

        auto [a,b] = ipd(pa,qa,pb,qb,n);
        results[i][0] += a.value / 4;               // remember that value samples are obtained for all directional derivatives,
        results[i][1] += b.value / 4;               // i.e. four times the number of directional derivative samples. Compensate here.
        results[i][2+2*d] += a.derivativeSample();
        results[i][3+2*d] += b.derivativeSample();

if (std::abs(a.derivativeSample()) > 1e6)
  std::cout << "d=" << d << " r=" << a.derivativeSample() << " sp.factor=" << sp.factor << "\n";
      }
    }
  };

  std::vector<std::future<void>> handles;
  for (int i=0; i<8; ++i)
    handles.push_back( std::async(std::launch::async,evaluate,i*N/8,(i+1)*N/8) );
  for (auto& f: handles)
    f.wait();


  // Compute mean
  for (auto const& r: results)
    for (int i=0; i<r.size(); ++i)
      mean[i] += r[i];
  for (auto& s: mean)                       // average out for obtaining the mean
    s /= N;

  // Compute sample variance
  for (auto const& r: results)
    for (int i=0; i<r.size(); ++i)
    {
      variance[i] += (r[i]-mean[i])*(r[i]-mean[i]);

if (std::abs(r[i]-mean[i]) > 1e6)
  std::cout << "i=" << i << " r[i]=" << r[i] << " mean[i]=" << mean[i] << "\n";
    }
  for (auto& s: variance)
    s /= N-1;

  return {mean,variance};
}

// ------------------------------------------------------------------------------------------------

int main(void)
{
  std::array rounds{1,3,9};

  // Obtain sample variance for different jump sampling probabilities.
  std::cout << "Computing sample variances\n";
  double h = 1./20;
  std::ofstream out("variance.gnu");
  for (double w=h/2; w<1; w+=h)
  {
    out << w;
    for (int n: rounds)
    {
      auto rs = runIpd(0.5,0.5,n,n*n*1000,w);
      out << ' ' << 4*rs[1][0] << ' ' << rs[1][2] << ' ' << rs[1][4];
if (n==1)
  std::cout << "da/dqa = " << rs[0][4] << "  var=" << rs[1][4] << "\n";
    }
    out << "\n";
  }

return 0;

  // Create cartesian grid evaluation of payoffs and their derivatives
  for (int n: rounds)
  {
    std::cout << "Computing for " << n << " rounds of game\n";
    std::ofstream out("out"+std::to_string(n)+".gnu");
    int N = n*n*1000;     // number of samples to draw for each strategy point


    double h = 1./20;
    for (double p=h/2; p<1; p+=h)
    {
      for (double q=h/2; q<1; q+=h)
      {
        auto rs = runIpd(p,q,n,N,0.5);
        out << p << " " << q;
        for (double r: rs[0])
          out << " " << r;
        out << "\n";
      }
      out << "\n";
    }
  }


  return 0;
}
