// Copyright 2023 Zuse Institute Berlin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "sad.hh"

#include <cassert>
#include <iostream>
#include <numeric>
#include <vector>

//-------------------------------------------------------------------------------------------------

namespace StochasticDifferentiation
{

  template <class Real>
  SAD<Real> bernoulli(SAD<Real> const& p)
  {
    assert(p.sp);

    Real om = p.sp->omega();
    Real value = om<p.value? 1: 0;

    if (p.sp->jumpMode)                               // jump mode: just propagate both values
    {                                                 //            using the same random omega drawn
      Real a = om < p.valA? 1: 0;
      Real b = om < p.valB? 1: 0;
      return SAD<Real>(p.sp,
                      value,
                      a-b,
                      a,b);
    }
    else
    {
      // So far, we've been in continuous mode. But Bernoulli distribution is discrete and
      // thus has jumps that we need to sample as well. The derivative is the sum of the
      // continuous and the jump contribution. We could sample them both with equal probability,
      // and weight both contributions with factor 2. But we can also weight them with w and
      // 1-w, respectively, and corresponding weight factors 1/w and 1/(1-w).
      Real w = p.sp->jumpWeight;
      if (p.sp->omega() < w)  // evaluate jump contribution with probability w
      {
        Real d_eta = 0<p.value && p.value<1? p.derivative: 0;  // jump location shifts with p in ]0,1[
        p.sp->updateWeight(d_eta / w);  // use weight 1/w
        p.sp->jumpMode = true;
        Real a = 1;
        Real b = 0;
        Real derivative = a-b;
        return SAD<Real>(p.sp,value,derivative,a,b);
      }
      else                    // evaluate continuous contribution with probability 1-w
      {
        p.sp->updateWeight(1/(1-w));
        Real derivative = 0;
        return SAD<Real>(p.sp,value,derivative);
      }
    }
  }

  template SAD<double> bernoulli(SAD<double> const& p);
  template SAD<float> bernoulli(SAD<float> const& p);

  //-------------------------------------------------------------------------------------------------

  template <class Sequence>
  typename Sequence::value_type generalizedBernoulli(Sequence const& ps)
  {
    using Real = typename Sequence::value_type::Real;

    // There must be at least one outcome.
    int const n = std::size(ps);
    assert(n > 0);

    // Make sure there is an active stochastic program.
    auto* sp = std::begin(ps)->sp;
    assert(sp);

    // The generalized Bernoulli distribution can be computed by drawing a uniformly distributed
    // omega and selecting the suitable interval from the cumulative distribution function. First
    // compute this one.
    std::vector<SAD<Real>> cdf(n,SAD<Real>(sp,0.0,0.0));
    std::partial_sum(begin(ps),end(ps),begin(cdf));

    // Make sure the probabilities add up exactly to 1, such that we don't fall off the end
    // with lower_bound below.
    assert(std::abs(1.0-Real(cdf.back())) < 10*std::numeric_limits<Real>::epsilon());
    cdf.back().value = std::max(Real(1.0),cdf.back().value);


    // Draw a real number omega in [0,1] and find the cdf interval in which it lies.
    // We do this as SAD<Real> for omega to be comparable within lower_bound.
    // We need to find i such that omega is in ]cdf_{i-1},cdf_i], with cdf_{-1} = 0
    // implicit. lower_bound returns an iterator pointing to the first cdf entry a such that
    // a >= omega, and therefore to cdf_i.
    SAD<Real> omega(sp,sp->omega(),0.0);
    int idx = std::lower_bound(begin(cdf),end(cdf),omega) - begin(cdf);
    assert(idx < n);

    Real value = idx;

    if (sp->jumpMode)     // already in jump mode: propagate both sides on
    {
      auto compA = [](auto const& a, auto const& b) { return a.valA < b.valA; };
      int idxA = std::lower_bound(begin(cdf),end(cdf),omega,compA) - begin(cdf);
      assert(idxA < n);

      auto compB = [](auto const& a, auto const& b) { return a.valB < b.valB; };
      int idxB = std::lower_bound(begin(cdf),end(cdf),omega,compB) - begin(cdf);
      assert(idxB < n);

      return SAD<Real>(sp,
                       value,
                       idxA-idxB,
                       idxA,idxB);
    }
    else                  // not yet in jump mode
    {
      // So far, we've been in continuous mode. But generalized Bernoulli distribution is discrete
      // and thus has jumps that we need to sample as well. The derivative is the sum of the
      // continuous and the jump contribution. We could sample them both with equal probability,
      // and weight both contributions with factor 2. But we can also weight them with w and
      // 1-w, respectively, and corresponding weight factors 1/w and 1/(1-w).
      Real w = sp->jumpWeight;
      if (sp->omega() < w)  // evaluate jump contribution with probability w
      {
        // There are n-1 jumps that enter the derivative, and we select one of them (randomly,
        // using the same omega that has already been used for the value in the first place).
        // This gives a jump index in the range 0,...,n-2.
        int idx = std::min(int((n-1)*Real(omega)),n-2);       // make sure we don't end up at n-1

        // We thus need to multiply the stochastic weight factor by n-1 in addition to
        // the jump location derivative. The jump location is just the cdf entry.
        Real d_eta = cdf[idx].derivative;
        sp->updateWeight(d_eta * (n-1) / w);
        sp->jumpMode = true;

        Real a = idx;
        Real b = idx+1;

        Real derivative = a-b;

        return SAD<Real>(sp,value,derivative,a,b);
      }
      else                    // evaluate continuous contribution with probability 1-w
      {
        sp->updateWeight(1/(1-w));
        Real derivative = 0;
        return SAD<Real>(sp,value,derivative);
      }
    }
  }

  template SAD<double> generalizedBernoulli(std::vector<SAD<double>> const& ps);
  template SAD<float> generalizedBernoulli(std::vector<SAD<float>> const& ps);

  //-------------------------------------------------------------------------------------------------

  template <class Real>
  SAD<Real> binomial(SAD<Real> const& p, int n, bool importanceSampling)
  {
    assert(p.sp);
    assert(n > 0);
    assert(0<=p.value && p.value<=1);

    // This computes the cumulative distribution function F of the binomial distribution at the
    // integer values k \in {0,...,n}:
    // F(k) = \sum_{j=0}^{k} \binom{n}{j} p^j (1-p)^{n-j}
    // We can extend this to the real interval omega \in [0,n] as F(omega) = F(\lceil omega \rceil).
    //
    // The inverse CDF consequently assumes the value k on the interval
    // ]F(k-1),F(k)] for k>=0. It jumps at the n positions
    // omega_k \in {F(0),...,F(n-1)} from k to k+1.
    //
    // Given k \in {0,...,n-1}, we compute the jump position omega_k = F(k) as well as its
    // derivative wrt p (which is always negative). In addition, we return the sum of all
    // derivatives (also negative), in order to facilitate importance sampling of jumps.
    auto distributionFunction = [n](int k, Real p)
    {
      assert(0<=k && k<n);

      if (p<0)
        return std::array<Real,2>{0.0,0.0};

      if (p>=1)
        return std::array<Real,2>{1.0,0.0};

      // Since omega_k = F(k) = \sum_{j=0}^k \binom{n}{j} p^j (1-p)^j,
      // we need to add up the interval lengths t_j = \binom{n}{j} p^j (1-p)^j.
      // By simple calculation, we get for t_0 = (1-p)^n and
      // t_{j+1} / t_j = (n-j)/(j+1) p/(1-p), such that we can iteratively compute t_j
      // and their derivatives.

      Real t = std::pow(1-p,n);         // t_0 (following jump 0)
      Real dt = -n*std::pow(1-p,n-1);   // and its derivative wrt p
      Real omega = t;                   // the position F(0) of jump k=1
      Real domega = dt;

      for (int j=0; j<k; ++j)
      {
        Real tfac = (n-j)/(j+1.0) * p/(1-p);              // the ratio between t_j and t_{j+1}
        Real dtfac = (n-j)/(j+1.0) * 1/( (1-p)*(1-p) );   // and its derivative wrt p
        dt = t*dtfac + dt*tfac;                           // derivative of t_{j+1} by product rule
        t *= tfac;                                        // t_{j+1} itself

        omega += t;                                       // the position F(j+1) of jump k=j+2
        domega += dt;                                     // and its derivative
      }

      return std::array<Real,2>{omega,domega};
    };

    auto sampleJump = [n](Real p, Real om)
    {
      assert(0<=om && om <=1);

      std::vector<Real> dOmega(n);

      Real t = std::pow(1-p,n);         // t_0 (following jump 0)
      Real dt = -n*std::pow(1-p,n-1);   // and its derivative wrt p
      Real omega = t;                   // the position F(0) of jump k=1
      Real domega = dt;

      for (int j=0; j<n; ++j)
      {
        dOmega[j] = domega;

        Real tfac = (n-j)/(j+1.0) * p/(1-p);              // the ratio between t_j and t_{j+1}
        Real dtfac = (n-j)/(j+1.0) * 1/( (1-p)*(1-p) );   // and its derivative wrt p
        dt = t*dtfac + dt*tfac;                           // derivative of t_{j+1} by product rule
        t *= tfac;                                        // t_{j+1} itself

        omega += t;                                       // the position F(j+1) of jump k=j+2
        domega += dt;                                     // and its derivative
      }

      std::vector<Real> dOmegaSums(n);
      std::partial_sum(begin(dOmega),end(dOmega),begin(dOmegaSums));
      Real total = dOmegaSums.back();
      for (Real& s: dOmegaSums)
        s /= total;
      assert(dOmegaSums.back()==1);
      auto it = std::lower_bound(begin(dOmegaSums),end(dOmegaSums),om);
      int k = it - begin(dOmegaSums);
      return std::pair(k,-total);
    };


    // This computes the inverse cumulative distribution function
    // F^{-1}(omega) = min {x \in [0,n] | F(x) >= omega}. Since F is constant
    // on intervals [k,k+1[, we obtain F^{-1}(omega) = min { k \in {0,...n} | F(k) >= omega}.
    // We use this to sample from the binomial distribution.
    auto inverseDistributionFunction = [n](Real omega, Real p)
    {
      if (omega <= 0)
        return -1;

      Real t = std::pow(1-p,n);
      Real F = t;
      for (int k=0; k<=n; ++k)
      {
        if (omega <= F)
          return k;

        t *= (n-k)/(k+1.0) * p/(1-p);
        F += t;
      }
      return n;
    };


    Real om = p.sp->omega();
    Real value = inverseDistributionFunction(om,p.value);
    Real derivative;
    Real a, b;

    if (p.sp->jumpMode)                                 // already in jump mode: just propagate
    {                                                   // both values with the same random omega
      a = inverseDistributionFunction(om,p.valA);
      b = inverseDistributionFunction(om,p.valB);
      return SAD<Real>(p.sp,
                       value,
                       a-b,
                       a,b);
    }
    else
    {
      Real w = p.sp->jumpWeight;
      if (p.sp->omega() < w)                            // evaluate jump contribution with probability w
      {
        int k;                                          // the (randomly) selected jump index

        if (importanceSampling)
        {
          // There are n jumps to choose from, with very different values of jump position derivatives
          // deta/dp (all negative, though). We perform importance sampling according to the absolute
          // values of deta/dp. Then the stochastic weight of the sample is just the sum of all
          // derivatives (absolute value, of course), and of course divided by the jump weight w.
          auto [kloc,total] = sampleJump(p.value,p.sp->omega());
          k = kloc;
          total *= p.derivative;                        // chain rule
          p.sp->updateWeight(total/w);
        }
        else
        {
          // There are n jumps to choose from, all with equal probability (because all have the same
          // jump height, which is 1). Select a sample k in {0,...,n-1}, which means the jump is from
          // k to k+1. As we choose one out of n possible jumps, the sample gets the weight n.
          int kloc = p.sp->omega()*n;
          k = std::min(kloc,n);

          // Then evaluate this jump's position eta and its derivative deta w.r.t. p. (Remember deta/dp
          // is negative, so we take its absolute value for the weight).
          auto [eta,deta] = distributionFunction(k,p.value);
          deta *= p.derivative;                // chain rule
          p.sp->updateWeight(-deta * n / w);   // use weight n/w
        }


        p.sp->jumpMode = true;
        a = k+1;
        b = k;
        derivative = 1;
        return SAD<Real>(p.sp,value,derivative,a,b);
      }
      else
      {
        p.sp->updateWeight(1/(1-w));    // use weight 1/(1-w)
        derivative = 0;
        return SAD<Real>(p.sp,value,derivative);
      }
    }
  }

  // explicit instantiation
  template  SAD<double> binomial(SAD<double> const& p, int n, bool importanceSampling);
  template  SAD<float> binomial(SAD<float> const& p, int n, bool importanceSampling);

  //-------------------------------------------------------------------------------------------------

  template <class Real>
  SAD<Real> gauss(SAD<Real> const& mu, SAD<Real> const& s)
  {
    auto sp = stochasticProgram(mu,s);
    Real om = sp->omega();
    Real normalSample = boost::math::erf_inv<Real>(2*om-1);
    Real value = normalSample*s.value+mu.value;

    if (sp->jumpMode)
    {
      Real a = normalSample*s.valA+mu.valA;
      Real b = normalSample*s.valB+mu.valB;
      Real derivative = a-b;
      return SAD<Real>(sp,value,derivative,a,b);
    }
    else
    {
      // There's no jump in the normal distribution function, hence we don't need
      // to distinguish between jump and continuous parts.
      Real derivative = normalSample*s.derivative + mu.derivative;
      return SAD<Real>(sp,value,derivative);
    }
  }

  // explicit instantiation
  template SAD<double> gauss(SAD<double> const& mu, SAD<double> const& s);
  template SAD<float> gauss(SAD<float> const& mu, SAD<float> const& s);

  //-------------------------------------------------------------------------------------------------

  template <class Real>
  SAD<Real> exponential(SAD<Real> const& lambda)
  {
    auto sp = lambda.sp;
    assert(sp);

    assert(Real(lambda) > 0);

    // Obtain a sample by drawing omega uniformly distributed in [0,1[ and
    // solving for the cumulative density function omega = 1-exp(-lambda*x).
    Real omega = sp->omega();
    while (omega == 1)          // This should be *very* rare, but it can happen...
      omega = sp->omega();
    Real y = -std::log(1-omega);
    Real x = y/Real(lambda);

    if (sp->jumpMode)             // we are in jump mode - just propagate the values on both sides
    {
      assert(lambda.valA>0 && lambda.valB>0);
      Real a = y/lambda.valA;
      Real b = y/lambda.valB;
      Real derivative = a-b;
      return SAD<Real>(sp,x,derivative,a,b);
    }
    else                          // not (yet) in jump mode - take continuous derivative
    {                             // dx/dlambda
      Real derivative = -y*lambda.derivative/Real(lambda)/Real(lambda);
      return SAD<Real>(sp,x,derivative);
    }
  }

  // explicit instantiation
  template SAD<double> exponential(SAD<double> const& lambda);
  template SAD<float> exponential(SAD<float> const& lambda);

}
