// Copyright 2023 Zuse Institute Berlin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SAD_HH
#define SAD_HH

#include "boost/math/special_functions/erf.hpp"

#include <cassert>
#include <memory>
#include <random>
#include <variant>


// ------------------------------------------------------------------------------------------------

namespace StochasticDifferentiation
{

  /**
   * \brief Maintains global state of a stochastic program derivative sample evaluation.
   */
  template <class Real>
  class StochasticProgram
  {
  public:
    bool jumpMode = false;

    Real jumpWeight = 0.5;

    /**
     * \brief Constructor.
     */
    StochasticProgram()
    : gen(std::random_device{}())
    , distOmega(0, 1)
    {
    }

    /**
     * \brief Returns a random number uniformly distributed in [0,1].
     */
    double omega()
    {
      return distOmega(gen);
    }

    /**
     * \brief The current stochastic weight of the sample.
     *
     * If an elementary function of a stochastic program is written as a sum of
     * two contributions to be sampled, e.g., a continuous part and a jump part,
     * then instead of sampling both (which could lead to a combinatorial explosion
     * of work), we sample just one and increase its stochastic weight by a factor
     * of two (if both contributions are chosen with equal probability 1/2).
     */
    Real weight() const
    {
      return factor;
    }

    /**
     * \brief updates the weight factor for the sample by the given factor w.
     *
     *
     * \param w the weight factor, must be at least 1.
     */
    void updateWeight(Real w)
    {
      factor *= w;
    }

    /**
     * \brief Returns a random number generator.
     */
    std::mt19937& generator()
    {
      return gen;
    }

  private:
    std::mt19937 gen;
    std::uniform_real_distribution<Real> distOmega;
    Real factor = 1;
  };

  // ------------------------------------------------------------------------------------------------

  /**
   * \brief A real arithmetic type supporting the sampling of stochastic program derivatives.
   */
  template <class Real_>
  class SAD
  {
    using Self = SAD<Real_>;

  public:
    /**
     * \brief Underlying real arithmetic type
     */
    using Real = Real_;

    /**
     * \brief Constructor
     */
    SAD(StochasticProgram<Real>* sp_, Real value_, Real derivative_)
    : sp(sp_)
    , value(value_)
    , derivative(derivative_)
    , valA(value)
    , valB(value)
    {
      assert(sp);
      factor = sp->weight();
    }

    /**
     * \brief Constructor
     */
    SAD(StochasticProgram<Real>* sp_, Real value_, Real derivative_, Real valA_, Real valB_)
    : sp(sp_)
    , value(value_)
    , derivative(derivative_)
    , valA(valA_)
    , valB(valB_)
    {
      assert(sp);
      assert(sp->jumpMode || (value== valA && valA==valB));
      factor = sp->weight();
    }

    /**
     * \brief Conversion to the underlying real type provides the sample value.
     */
    explicit operator Real() const
    {
      return value;
    }

    /**
     * \brief Returns the computed sample of the parametric derivative.
     */
    Real derivativeSample() const
    {
      return factor*derivative;
    }

    /**
     * \brief In-place addition.
     */
    Self& operator += (Self const& x)
    {
      *this = *this + x;
      return *this;
    }

    /**
     * \brief In-place subtraction.
     */
    Self& operator -= (Self const& x)
    {
      *this = *this - x;
      return *this;
    }

    // Low level direct access interface - don't use. Or only if you must
    // and know what you're doing.

    StochasticProgram<Real>* sp;
    Real value = 0;
    Real derivative = 0;
    Real valA = 0, valB = 0;
    Real factor;
  };

  // ------------------------------------------------------------------------------------------------

  /**
   * \brief Utility function that gets the stochastic program state of a real value and makes sure that
   *        both provided values belong to the same stochastic program evaluation.
   */
  template <class Real>
  StochasticProgram<Real>* stochasticProgram(SAD<Real> const& x, SAD<Real> const& y)
  {
    assert(x.sp==y.sp && x.sp);
    return x.sp;
  }

  // ------------------------------------------------------------------------------------------------

  /**
   * \defgroup arithmetic Arithmetic operations
   *
   * Basic arithmetic operations on samples of stochastic programs.
   */


  /**
   * \ingroup arithmetic
   * \brief Multiplication by a scalar constant
   * \relates SAD
   */
  template <class T, class Real, std::enable_if_t<std::is_arithmetic_v<T>,bool> = true>
  SAD<Real> operator*(T c_, SAD<Real> const& x)
  {
    Real c = Real(c_);
    Real deriv = x.sp->jumpMode? c*(x.valA-x.valB)
                               : c*x.derivative;

    return SAD<Real>(x.sp,c*x.value,
                     deriv,
                     c*x.valA,c*x.valB);
  }

  /**
   * \ingroup arithmetic
   * \brief Multiplication by a scalar constant
   * \relates SAD
   */
  template <class T, class Real>
  SAD<Real> operator*(SAD<Real> const& x, T c)
  {
    return c*x;
  }

 /**
   * \ingroup arithmetic
   * \brief Division by a scalar constant
   * \relates SAD
   */
  template <class T, class Real, std::enable_if_t<std::is_arithmetic_v<T>,bool> = true>
  SAD<Real> operator/(SAD<Real> const& x, T c_)
  {
    Real c = Real(c_);
    Real deriv = x.sp->jumpMode? (x.valA-x.valB)/c
                               : x.derivative/c;

    return SAD<Real>(x.sp,x.value/c,
                     deriv,
                     x.valA/c,x.valB/c);
  }


  /**
   * \ingroup arithmetic
   * \brief Addition of a scalar constant
   * \relates SAD
   */
  template <class T, class Real, std::enable_if_t<std::is_arithmetic_v<T>,bool> = true>
  SAD<Real> operator+(T c_, SAD<Real> const& x)
  {
    Real c = Real(c_);
    Real deriv = x.sp->jumpMode? x.valA-x.valB
                               : x.derivative;

    return SAD<Real>(x.sp,c+x.value,
                     deriv,
                     c+x.valA,c+x.valB);
  }

  /**
   * \ingroup arithmetic
   * \brief Addition of a scalar constant
   * \relates SAD
   */
  template <class T, class Real, std::enable_if_t<std::is_arithmetic_v<T>,bool> = true>
  SAD<Real> operator+(SAD<Real> const& x, T c)
  {
    return c+x;
  }

  /**
   * \ingroup arithmetic
   * \brief Subtraction from a scalar constant
   * \relates SAD
   */
  template <class T, class Real, std::enable_if_t<std::is_arithmetic_v<T>,bool> = true>
  SAD<Real> operator-(T c_, SAD<Real> const& x)
  {
    Real c = Real(c_);
    Real deriv = x.sp->jumpMode? -(x.valA-x.valB)
                               : -x.derivative;

    return SAD<Real>(x.sp,c-x.value,
                     deriv,
                     c-x.valA,c-x.valB);
  }

  /**
   * \ingroup arithmetic
   * \brief Subtraction of a scalar constant
   * \relates SAD
   */
  template <class Real>
  SAD<Real> operator-(SAD<Real> const& x, Real c)
  {
    return x + (-c);
  }

  // ------------------------------------------------------------------------------------------------

  /**
   * \ingroup arithmetic
   * \brief Negation of a real
   * \relates SAD
   */
  template <class Real>
  SAD<Real> operator-(SAD<Real> const& x)
  {
    return -1.0 * x; // shortest but possibly not most performant implementation
  }

  // ------------------------------------------------------------------------------------------------

  /**
   * \ingroup arithmetic
   * \brief Sum of two real samples
   * \relates SAD
   */
  template <class Real>
  SAD<Real> operator+(SAD<Real> const& x, SAD<Real> const& y)
  {
    auto sp = stochasticProgram(x,y);

    // If in jump mode, we just propagate forward A and B values independently of the value sample
    // itself. The derivative sample is evaluated as factor*(valA-valB).
    if (sp->jumpMode)
    {
      Real a = x.valA+y.valA;
      Real b = x.valB+y.valB;

      return SAD<Real>(sp,x.value+y.value,
                       a-b,
                       a,b);
    }
    // Otherwise, we propagate the value and store the same in A and B values (before jump mode,
    // A and B are supposed to agree with value).
    // In either case, we can use:
    assert(x.value==x.valA && x.value==x.valB && y.value==y.valA && y.value==y.valB);
    return SAD<Real>(sp,x.value+y.value,
                     x.derivative+y.derivative);
  }

  /**
   * \ingroup arithmetic
   * \brief Difference of two real samples
   * \relates SAD
   */
  template <class Real>
  SAD<Real> operator-(SAD<Real> const& x, SAD<Real> const& y)
  {
    auto sp = stochasticProgram(x,y);

    // If in jump mode, we just propagate forward A and B values independently of the value sample
    // itself. The derivative sample is evaluated as factor*(valA-valB).
    if (sp->jumpMode)
    {
      Real a = x.valA - y.valA;
      Real b = x.valB - y.valB;

      return SAD<Real>(sp,x.value-y.value,
                       a-b,
                       a,b);
    }
    // Otherwise, we propagate the value and store the same in A and B values (before jump mode,
    // A and B are supposed to agree with value).
    // In either case, we can use:
    assert(x.value==x.valA && x.value==x.valB && y.value==y.valA && y.value==y.valB);
    return SAD<Real>(sp,x.value-y.value,
                     x.derivative-y.derivative);
  }


  /**
   * \ingroup arithmetic
   * \brief Product of two real samples
   * \relates SAD
   */
  template <class Real>
  SAD<Real> operator*(SAD<Real> const& x, SAD<Real> const& y)
  {
    auto sp = stochasticProgram(x,y);

    // If in jump mode, we just propagate forward A and B values independently of the value sample
    // itself. The derivative sample is evaluated as factor*(valA-valB).
    if (sp->jumpMode)
    {
      Real a = x.valA * y.valA;
      Real b = x.valB * y.valB;

      return SAD<Real>(sp,x.value*y.value,
                           a-b,
                           a,b);
    }
    // Otherwise, we propagate the value and store the same in A and B values (before jump mode,
    // A and B are supposed to agree with value).
    // In either case, we can use:
    assert(x.value==x.valA && x.value==x.valB && y.value==y.valA && y.value==y.valB);
    return SAD<Real>(sp,x.value*y.value,
                         x.derivative*y.value+x.value*y.derivative);
  }


  /**
   * \ingroup arithmetic
   * \brief Quotient \f$ x/y \f$ of two real samples
   * \relates SAD
   */
  template <class Real>
  SAD<Real> operator/(SAD<Real> const& x, SAD<Real> const& y)
  {
    auto sp = stochasticProgram(x,y);

    // If in jump mode, we just propagate forward A and B values independently of the value sample
    // itself.
    if (sp->jumpMode)
    {
      Real a = x.valA / y.valA;
      Real b = x.valB / y.valB;

      return SAD<Real>(sp, x.value/y.value, a-b, a,b);
    }
    // Otherwise, we propagate the value and store the same in A and B values (before jump mode,
    // A and B are supposed to agree with value).
    // In either case, we can use:
    assert(x.value==x.valA && x.value==x.valB && y.value==y.valA && y.value==y.valB);
    return SAD<Real>(sp, x.value/y.value,
                         (x.derivative*y.value-x.value*y.derivative)/(y.value*y.value));
  }





  // ------------------------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------------------------

  /**
   * \defgroup comparison Comparison operations
   *
   * Ordering of real stochastic program samples based on their value.
   */

  /**
   * \ingroup comparison
   * \brief less
   */
  template <class Real>
  bool operator < (SAD<Real> const& a, SAD<Real> const& b)
  {
    return a.value < b.value;
  }

  /**
   * \ingroup comparison
   * \brief less
   */
  template <class Real>
  bool operator < (SAD<Real> const& a, Real b)
  {
    return a.value < b;
  }

  /**
   * \ingroup comparison
   * \brief less
   */
  template <class Real>
  bool operator < (Real a, SAD<Real> const& b)
  {
    return a < b.value;
  }


  // ------------------------------------------------------------------------------------------------
  // ------------------------------------------------------------------------------------------------

  /**
   * \defgroup distributions Continuous and discontinuous probability distributions
   *
   * These are parameterized probability distributions which draw one sample on each call.
   * The samples can be differentiated (in mean) with respect to the distribution's parameters.
   */

  /**
   * \ingroup distributions
   * \brief The Bernoulli distribution with success probability \f$ p \f$.
   *
   * \param[in] p the success probability
   *
   * This returns 1 with probability \f$ p \f$ and 0 with probability \f$ 1-p \f$. Here,
   * \f$ p \f$ is interpreted as \f$ \max(0,\min(p,1)) \f$.
   *
   * \relates SAD
   */
  template <class Real>
  SAD<Real> bernoulli(SAD<Real> const& p);

  // ------------------------------------------------------------------------------------------------

  /**
   * \ingroup distributions
   * \brief The generalized Bernoulli distribution.
   *
   * This returns an integer \f$ i \f$ in the range \f$ 0, \dots, n-1 \f$ with
   * probability \f$ p_i \f$.
   *
   * \tparam Sequence an STL container with value type SAD<Real>
   *
   * \param[in] ps the success probabilities. Precondition: ps is contained in the probability simplex
   *               characterized by  \f$ \sum_{i=0}^{n-1} p_i = 1 \wedge p_i \ge 0 \forall i\f$.
   * \relates SAD
   */
  template <class Sequence>
  typename Sequence::value_type generalizedBernoulli(Sequence const& ps);

  // ------------------------------------------------------------------------------------------------

  /**
   * \ingroup distributions
   * \brief The binomial distribution with success probability p and n rounds.
   *
   * \param p success probability in each round
   * \param n number of rounds
   * \param importanceSampling whether to perform importance sampling for jumps
   *
   * When sampling jump contributions, there are \f$ n \f$ different jumps, with
   * (depending on \f$ n \f$ very) different values for the jump position derivative wrt \f$ p \f$.
   * Since this derivative ends up in the stochastic weight, the pronounced non-uniformity
   * induces a high variance. This can be mitigated by importance sampling, i.e. choosing
   * a jump with probability proportional to the position derivative.
   *
   * The drawback of importance sampling is its higher memory demand \f$ \mathcal{O}(n) \f$,
   * also incurring memory allocation, and somewhat higher compute load (roughly a factor 2).
   * \relates SAD
   */
  template <class Real>
  SAD<Real> binomial(SAD<Real> const& p, int n, bool importanceSampling=true);

  // ------------------------------------------------------------------------------------------------

  /**
   * \ingroup distributions
   * \brief The Gaussian distribution with mean \f$ \mu \f$  and standard deviation \f$ \sigma \f$.
   * \relates SAD
   */
  template <class Real>
  SAD<Real> gauss(SAD<Real> const& mu, SAD<Real> const& sigma);

  // ------------------------------------------------------------------------------------------------

  /**
   * \ingroup distributions
   * \brief The exponential distribution with parameter \f$ \lambda \f$.
   *
   * This draws a sample \f$ x \f$ distributed exponentially, i.e.
   * according to  \f[ f_\lambda(x) =
   * \begin{cases} \lambda e^{-\lambda x}, & x \ge 0 \\
   *               0, & x < 0 \end{cases}. \f]
   * The mean is \f$ 1/\lambda \f$.
   * \relates SAD
   */
  template <class Real>
  SAD<Real> exponential(SAD<Real> const& lambda);

}


#endif
