// Copyright 2023 Zuse Institute Berlin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


// Test program for differentiating binomial distributions.


#include <array>
#include <cmath>
#include <iostream>
#include <tuple>

#include <boost/program_options.hpp>

#include "sad.hh"

auto sqr = [](double x) { return x*x; };


using namespace StochasticDifferentiation;

// We define a stochastic program that draws a sample from the
// binomial distribution with n trials and success probability p.
// This tests the "not yet in jump mode" branch of the binomial
// distribution.
//
// The mean is np, and its derivative therefore n.
//
//
// We return a tuple (a,b,A,B), where a is a sample of \pi, b a sample
// of the mean's derivative w.r.t. p, A is the mean of \pi, and B
// its derivative wrt. p, i.e. A and B are the exact values of which
// a and b are unbiased samples.
std::array<double,4> getSample1(double p_, int n, bool importance)
{
  using Real = SAD<double>;

  StochasticProgram<double> sp;
  sp.jumpWeight = 0.5;

  Real p(&sp,p_,1.0);
  Real x = binomial(p,n,importance);


  double mean = n*p_;
  double dmean = n;

  return {double(x),x.derivativeSample(),
          mean, dmean };
}

// We define a stochastic program that draws a sample from the
// composite distribution Bin(z,n) with z = 1/4+Ber(p)/2. This tests the "jump mode"
// branch of the binomial distribution.
//
// The mean of Bin(z) is nz. z is 3/4 with probability p and 1/4 with probability
// 1-p. The total mean is therefore 3pn/4 + (1-p)n/4 = pn/2 + n/4, with derivative n/2.
//
//
// We return a tuple (a,b,A,B), where a is a sample of \pi, b a sample
// of the mean's derivative w.r.t. p, A is the mean of \pi, and B
// its derivative wrt. p, i.e. A and B are the exact values of which
// a and b are unbiased samples.
std::array<double,4> getSample2(double p_, int n, bool importance)
{
  using Real = SAD<double>;

  StochasticProgram<double> sp;
  sp.jumpWeight = 0.5;

  Real p(&sp,p_,1.0);
  Real z = 1.0/4 + bernoulli(p)/2;
  Real x = binomial(z,n,importance);

  double mean = p_*n/2+n/4.0;
  double dmean = n/2.0;

  return {double(x),x.derivativeSample(),
          mean, dmean };
}

// Compute the mean of the distribution above and the mean's derivative
// up to an estimated Monte Carlo error of tol.
std::tuple<int,double,double> checkFor(double p, int n, bool importance, double tol,
                                       std::array<double,4> (*getSample)(double p_, int n, bool importance),
                                       bool verbose)
{
  double piSum=0, dpiSum=0;
  double dpiSumFD=0;
  double piVarSum=0, dpiVarSum=0;

  double err = 2*tol+1;
  int i = 0;
  int imax = 1000000;
  for ( ; err>=tol && i<imax; ++i)
  {
    double h = 0.2;
    auto [a,b,A,B] = (*getSample)(p,n,importance);
    auto [am,bm,Am,Bm] = (*getSample)(p-h/2,n,importance);
    auto [a_,b_,A_,B_] = (*getSample)(p+h/2,n,importance);
    piSum += a;
    dpiSum += b;
    dpiSumFD += (a_-am)/h;


    piVarSum += sqr(a-A);
    dpiVarSum += sqr(b-B);

    double sampleVariance = std::max(piVarSum,dpiVarSum)/(i+1);
    err = std::sqrt(sampleVariance/(i+1));

    if (verbose && i%10000 == 0)
      std::cerr << "\n" << "         pi=" << piSum/(i+1) << " (-> " << A << "),  dpi="
                << dpiSum/(i+1) << " (-> " << B << " or FD " << dpiSumFD/(i+1) << ")  err=" << err << " vs tol=" << tol;
  }

  if (i==imax)  // notify that we didn't reach the required accuracy within the iteration limit
    i = -i;

  return {i,piVarSum/(i+1),dpiVarSum/(i+1)};
}


bool runTest(std::string const& title, int n, bool importance, double tol,
             std::array<double,4> (*getSample)(double p_, int n, bool importance),
             bool verbose)
{
  bool success = true;

  std::cout << title; std::cout.flush();
  for (double p=0.2; p<0.8; p+=0.1)
  {
    double tol = 3e-2;
    auto [i,piVar,dpiVar] = checkFor(p,n,importance,tol,getSample1,verbose);
    success = success && i>=0;
    if (verbose)
      std::cout << "\np = " << p << ": i=" << i << " piVariance=" << piVar << " dpiVariance=" << dpiVar << " -> "
                << (i>0? "pass": "fail") << "\n";
    else
      std::cout << ".";
    std::cout.flush();
  }

  return success;
}


int main(int argc, char** argv)
{
  namespace po = boost::program_options;
  po::options_description desc("Allowed options");
  desc.add_options()
    ("n",       po::value<int>()->default_value(8),     "number of Bernoulli trials")
    ("verbose", po::value<bool>()->default_value(true), "verbose output");
  po::variables_map vm;
  po::store(po::parse_command_line(argc,argv,desc),vm);
  po::notify(vm);


  bool verbose = vm["verbose"].as<bool>();

  // Note that for larger n the variance is higher, such that larger tolerances should be used
  // (or the tests fail).
  int n = vm["n"].as<int>();

  bool success = true;
  success = runTest("Testing binomial continuous mode ",n,false,3e-2,getSample1,verbose)                        && success;
  success = runTest("Testing binomial jump mode ",n,false,1e-2,getSample1,verbose)                              && success;
  success = runTest("Testing binomial continuous mode (importance sampling) ",n,true,3e-2,getSample1,verbose)   && success;
  success = runTest("Testing binomial jump mode (importance sampling) ",n,true,1e-2,getSample1,verbose)         && success;


  std::cout << "\nTest " << (success? "succeeded": "failed") << ".\n";
  return success? 0: -1;
}
