// Copyright 2023 Zuse Institute Berlin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


// Test program for differentiating exponential distributions.


#include <array>
#include <cmath>
#include <iostream>
#include <tuple>

#include <boost/program_options.hpp>

#include "sad.hh"

auto sqr = [](double x) { return x*x; };


using namespace StochasticDifferentiation;

// We define a stochastic program that draws a sample from the
// exponential distribution with mean 1/p.
// This tests the "not yet in jump mode" branch of the exponential
// distribution.
//
// The mean is 1/p, and its derivative therefore -1/p^2.
//
//
// We return a tuple (a,b,A,B), where a is a sample of \pi, b a sample
// of the mean's derivative w.r.t. p, A is the mean of \pi, and B
// its derivative wrt. p, i.e. A and B are the exact values of which
// a and b are unbiased samples.
std::array<double,4> getSample1(double p_)
{
  using Real = SAD<double>;

  StochasticProgram<double> sp;
  sp.jumpWeight = 0.5;

  Real p(&sp,p_,1.0);
  Real x = exponential(p);


  double mean = 1/p_;
  double dmean = -1/sqr(p_);

  return {double(x),x.derivativeSample(),
          mean, dmean };
}

// We define a stochastic program that draws a sample from the
// composite distribution Exp(z) with z = 1+Ber(p). This tests the "jump mode"
// branch of the exponential distribution.
//
// The mean of Exp(z) is 1/z. z is 2 with probability p and 1 with probability
// 1-p. The total mean is therefore p/2 + (1-p) = 1-p/2, with derivative -1/2.
//
//
// We return a tuple (a,b,A,B), where a is a sample of \pi, b a sample
// of the mean's derivative w.r.t. p, A is the mean of \pi, and B
// its derivative wrt. p, i.e. A and B are the exact values of which
// a and b are unbiased samples.
std::array<double,4> getSample2(double p_)
{
  using Real = SAD<double>;

  StochasticProgram<double> sp;
  sp.jumpWeight = 0.5;

  Real p(&sp,p_,1.0);
  Real z = 1 + bernoulli(p);
  Real x = exponential(z);

  double mean = 1-p_/2;
  double dmean = -0.5;

  return {double(x),x.derivativeSample(),
          mean, dmean };
}

// Compute the mean of the distribution above and the mean's derivative
// up to an estimated Monte Carlo error of tol.
std::tuple<int,double,double> checkFor(double p, double tol,
                                       std::array<double,4> (*getSample)(double p_),
                                       bool verbose)
{
  double piSum=0, dpiSum=0;
  double dpiSumFD=0;
  double piVarSum=0, dpiVarSum=0;

  double err = 2*tol+1;
  int i = 0;
  int imax = 1000000;
  for ( ; err>=tol && i<imax; ++i)
  {
    double h = 0.2;
    auto [a,b,A,B] = (*getSample)(p);
    auto [am,bm,Am,Bm] = (*getSample)(p-h/2);
    auto [a_,b_,A_,B_] = (*getSample)(p+h/2);
    piSum += a;
    dpiSum += b;
    dpiSumFD += (a_-am)/h;


    piVarSum += sqr(a-A);
    dpiVarSum += sqr(b-B);

    double sampleVariance = std::max(piVarSum,dpiVarSum)/(i+1);
    err = std::sqrt(sampleVariance/(i+1));

    if (verbose && i%10000 == 0)
      std::cerr << "\n" << "         pi=" << piSum/(i+1) << " (-> " << A << "),  dpi="
                << dpiSum/(i+1) << " (-> " << B << " or FD " << dpiSumFD/(i+1) << ")  err=" << err << " vs tol=" << tol;
  }

  if (i==imax)  // notify that we didn't reach the required accuracy within the iteration limit
    i = -i;

  return {i,piVarSum/(i+1),dpiVarSum/(i+1)};
}


int main(int argc, char** argv)
{
  namespace po = boost::program_options;
  po::options_description desc("Allowed options");
  desc.add_options()
    ("verbose", po::value<bool>()->default_value(true), "verbose output");
  po::variables_map vm;
  po::store(po::parse_command_line(argc,argv,desc),vm);
  po::notify(vm);


  bool verbose = vm["verbose"].as<bool>();
  bool success = true;

  std::cout << "Testing exponential continuous mode "; std::cout.flush();
  for (double p=0.2; p<0.8; p+=0.1)
  {
    double tol = 3e-2;
    auto [i,piVar,dpiVar] = checkFor(p,tol,getSample1,verbose);
    success = success && i>=0;
    if (verbose)
      std::cout << "\np = " << p << ": i=" << i << " piVariance=" << piVar << " dpiVariance=" << dpiVar << "\n";
    else
      std::cout << ".";
    std::cout.flush();
  }

  std::cout << "\nTesting exponential jump mode "; std::cout.flush();
  for (double p=0.2; p<0.8; p+=0.1)
  {
    double tol = 5e-3;
    auto [i,piVar,dpiVar] = checkFor(p,tol,getSample2,verbose);
    success = success && i>=0;
    if (verbose)
      std::cout << "\np = " << p << ": i=" << i << " piVariance=" << piVar << " dpiVariance=" << dpiVar << "\n";
    else
      std::cout << ".";
    std::cout.flush();
  }

  std::cout << "\nTest " << (success? "succeeded": "failed") << ".\n";
  return success? 0: -1;
}
