// Copyright 2023 Zuse Institute Berlin
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


// Test program for differentiating generalized Bernoulli distributions.


#include <array>
#include <cmath>
#include <iostream>
#include <tuple>

#include "sad.hh"

auto sqr = [](double x) { return x*x; };


using namespace StochasticDifferentiation;

// We define a stochastic program that draws a sample from the
// distribution \pi = GBer(1-p,p/(n-1),...,p/(n-1)).
// This tests the "not yet in jump mode" branch of the generalized
// Bernoulli distribution.
//
// The mean of GBer(p) is p/(n-1) sum(i=1..n-1,i) = p/(n-1) n(n-1)/2
// = pn/2, and its derivative therefore n/2.
//
//
// We return a tuple (a,b,A,B), where a is a sample of \pi, b a sample
// of the mean's derivative w.r.t. p, A is the mean of \pi, and B
// its derivative wrt. p, i.e. A and B are the exact values of which
// a and b are unbiased samples.
std::array<double,4> getSample1(int n, double p_)
{
  using Real = SAD<double>;

  StochasticProgram<double> sp;
  sp.jumpWeight = 0.5;

  Real p(&sp,p_,1.0);
  std::vector<Real> ps(n,p/(n-1));
  ps[0] = 1-p;
  Real r = generalizedBernoulli(ps);


  double mean = p_*n/2;
  double dmean = n/2.;

  return {double(r),r.derivativeSample(),
          mean, dmean };
}

// We define a stochastic program that draws a sample from the
// composite distribution \pi = GBer(1-z,z/(n-1),...,z/(n-1))
// with z = 1/4+Ber(p)/2.  This tests the "jump mode" branch of the generalized
// Bernoulli distribution.
//
// The mean of GBer(z) is z/(n-1) sum(i=1..n-1,i) = z/(n-1) n(n-1)/2
// = zn/2. z is 3/4 with probability p and 1/4 with probability 1-p. The
// total mean is therefore 3pn/8 + (1-p)n/8 = 2pn/8 + n/8 = pn/4 + n/8,
// with a derivative n/4.
//
//
// We return a tuple (a,b,A,B), where a is a sample of \pi, b a sample
// of the mean's derivative w.r.t. p, A is the mean of \pi, and B
// its derivative wrt. p, i.e. A and B are the exact values of which
// a and b are unbiased samples.
std::array<double,4> getSample2(int n, double p_)
{
  using Real = SAD<double>;

  StochasticProgram<double> sp;
  sp.jumpWeight = 0.5;

  Real p(&sp,p_,1.0);
  Real z = 0.25 + bernoulli(p)/2;
  std::vector<Real> ps(n,z/(n-1));
  ps[0] = 1-z;
  Real r = generalizedBernoulli(ps);


  double mean = p_*n/4 + n/8.0;
  double dmean = n/4.0;

  return {double(r),r.derivativeSample(),
          mean, dmean };
}

// Compute the mean of the distribution above and the mean's derivative
// up to an estimated Monte Carlo error of tol.
std::tuple<int,double,double> checkFor(int n, double p, double tol,
                                       std::array<double,4> (*getSample)(int n, double p_))
{
  double piSum=0, dpiSum=0;
  double dpiSumFD=0;
  double piVarSum=0, dpiVarSum=0;

  double err = 2*tol+1;
  int i = 0;
  for ( ; err>=tol && i<100000; ++i)
  {
    double h = 0.2;
    auto [a,b,A,B] = (*getSample)(n,p);
    auto [am,bm,Am,Bm] = (*getSample)(n,p-h/2);
    auto [a_,b_,A_,B_] = (*getSample)(n,p+h/2);
    piSum += a;
    dpiSum += b;
    dpiSumFD += (a_-am)/h;


    piVarSum += sqr(a-A);
    dpiVarSum += sqr(b-B);

    double sampleVariance = std::max(piVarSum,dpiVarSum)/(i+1);
    err = std::sqrt(sampleVariance/(i+1));

    if (i%10000 == 0)
    std::cerr << "         pi=" << piSum/(i+1) << " (-> " << A << "),  dpi="
              << dpiSum/(i+1) << " (-> " << B << " or FD " << dpiSumFD/(i+1) << ")  err=" << err << " vs tol=" << tol << "\n";
  }

  return {i,piVarSum/(i+1),dpiVarSum/(i+1)};
}


int main(void)
{
  int n = 6;
  double tol = 1e-2;

  for (double p=0.1; p<0.8; p+=0.1)
  {
    auto [i,piVar,dpiVar] = checkFor(n,p,tol,getSample2);
    std::cout << "p = " << p << ": i=" << i << " piVar=" << piVar << " dpiVar=" << dpiVar << "\n";
  }
  return 0;
}
